//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef MICRODIMS_H
#define MICRODIMS_H 1

#include "TransProg/scheduler.h"
#include "TransProg/spatialUtils.h"
#include "TransProg/random.h"
#include "TransProg/path.h"
#include "TransProg/dynamicCompiler.h"
#include "TransCuda/cuda.h"

#define ENABLE_RENDERING 1

#define SYNCHRONISE_SUBSTRATE 0

#define ASSUMED_MAX_BACTERIUM_TYPE_COUNT 8 // increase as needed

//----------------------------------------------------------------------------

typedef struct _Bacterium Bacterium;
typedef struct _BacteriumStat BacteriumStat;
typedef struct _DetectionCellStack DetectionCellStack;
typedef struct _AcidRecord AcidRecord;

typedef
void
(*BacteriumCreationFnct)(int type,
                         real32_t x,
                         real32_t y,
                         real32_t mass,
                         DetectionCellStack * RESTRICT toStacks,
                         real32_t *toMasses,
                         int64_t nextStep,
                         int level,
                         int cpuCtx);

typedef
void
(*BacteriumRenderingFnct)(int cpuCtx,
                          int type,
                          real32_t x,
                          real32_t y,
                          real32_t radius,
                          int level,
                          real32_t growthAmount);

typedef
intptr_t // destination tile key
(*BacteriumActivationFnct)(const Bacterium * RESTRICT previousBacterium,
                           const BacteriumStat * RESTRICT bstat,
                           const DetectionCellStack * RESTRICT fromStacks,
                           DetectionCellStack * RESTRICT toStacks,
                           const real32_t * RESTRICT fromMasses,
                           real32_t *toMasses,
                           const real32_t * RESTRICT cG_array,
                           real32_t *delta_cG_array,
                           const real32_t * RESTRICT cL_array,
                           real32_t *delta_cL_array,
                           const real32_t * RESTRICT cA_array,
                           real32_t *delta_cA_array,
                           const real32_t * RESTRICT cF_array,
                           real32_t *delta_cF_array,
                           const AcidRecord * RESTRICT acidRecords,
                           int64_t currentStep,
                           int level,
                           int cpuCtx,
                           BacteriumRenderingFnct renderingFnct,
                           bool pause);

//----------------------------------------------------------------------------

struct _AcidRecord
{
CACHE_ALIGNED
real32_t maintCoef;      // [MG/(gDW.s)]
real32_t massYieldCoef;  // [gDW/MG]
real32_t qL_glucose;     // [MA/MG]
real32_t qL_maintenance; // [MA/(gDW.s)]
real32_t qA_glucose;     // [MA/MG]
real32_t qA_maintenance; // [MA/(gDW.s)]
real32_t qF_glucose;     // [MA/MG]
real32_t qF_maintenance; // [MA/(gDW.s)]
};

typedef struct _BacteriumConstantRecord
{
char *name;
//-- simulation specific parameters --
int initial_count;
real32_t distribution_center_x;
real32_t distribution_center_y;
real32_t distribution_inner_crown;
real32_t distribution_outer_crown;
//-- display specific parameters --
real32_t display_red;
real32_t display_green;
real32_t display_blue;
//-- model specific parameters --
real32_t shoving_factor;
real32_t bacterium_density;
real32_t k_m_glucose;
real32_t cP_min;
real32_t q_glucose_max;
real32_t lactate_acid_dissociation;
real32_t acetate_acid_dissociation;
real32_t formate_acid_dissociation;
real32_t u_min_lactate;
real32_t u_min_acetate;
real32_t u_min_formate;
real32_t m_pfl_glucose_ref;
real32_t m_ldh_glucose_ref;
real32_t q_lactate_no_growth;
real32_t q_acetate_no_growth;
real32_t q_formate_no_growth;
real32_t y_lactate;
real32_t y_acetate;
real32_t y_formate;
real32_t y_biomass_pfl;
real32_t y_biomass_ldh;
real32_t initiation_ratio_at_zero_growth;
real32_t initiation_ratio_slope;
real32_t replication_cycle_ps1;
real32_t replication_cycle_ps2;
real32_t replication_cycle_ps3;
//-- deduced from above --
real32_t mass_to_radius; // [m^3/gDW]
AcidRecord *acidRecords;
char *constantSourceCode;
PluginHandle dynamicCode;
BacteriumActivationFnct activationFnct;
BacteriumCreationFnct creationFnct;
} BacteriumConstantRecord;

typedef struct _ConstantRecord
{
//-- simulation specific parameters --
char *log_file_name;
real32_t log_period;
int maximal_bacterium_count;
real32_t maximal_simulation_duration;
real32_t initial_glucose_concentration;
char *acid_file_name;
int acid_max_step;
real32_t cL_max;
real32_t cA_max;
real32_t cF_max;
real32_t dt;
real32_t detection_step;
real32_t activation_step;
real32_t substrate_step;
real32_t substrate_size;
real32_t substrate_thickness;
real32_t substrate_plane_progression;
real32_t substrate_CFL;
int substrate_force_borders;
int substrate_crank_nicolson;
int substrate_iteration_count;
real32_t diffusivity_plot_width;
real32_t diffusivity_plot_amount;
real32_t diffusivity_plot_value;
//-- display specific parameters --
real32_t display_glucose_min;
real32_t display_glucose_max;
real32_t display_lactate_min;
real32_t display_lactate_max;
real32_t display_acetate_min;
real32_t display_acetate_max;
real32_t display_formate_min;
real32_t display_formate_max;
//-- model specific parameters for environment --
real32_t glucose_diffusivity;
real32_t lactate_diffusivity;
real32_t acetate_diffusivity;
real32_t formate_diffusivity;
//-- deduced from above --
int acid_step_count; // [1]
real32_t acid_scale; // [1]
real32_t div_cL_max; // [l/MA]
real32_t div_cA_max; // [l/MA]
real32_t div_cF_max; // [l/MA]
int detection_sampling; // [1]
real32_t detection_div_step; // [1/m]
real32_t detection_surface; // [m^2]
int activation_tiling; // [1]
int activation_sampling; // [1]
int substrate_sampling; // [1]
real32_t substrate_half_size; // [m]
real32_t substrate_loop; // [m]
real32_t substrate_div_step; // [1/m]
real32_t substrate_cell_volume; // [l]
int substrate_plane_count; // [1]
int substrate_dt_divisor; // [1]
real32_t radius_min; // [m]
real32_t radius_range; // [m]
int detection_cell_size; // [1]
int detection_cell_capacity; // [1]
//-- bacterium specific parameters --
char *constantSourceCode;
int bacterium_type_count;
BacteriumConstantRecord bacterium[ASSUMED_MAX_BACTERIUM_TYPE_COUNT];
} ConstantRecord;

//----------------------------------------------------------------------------

typedef enum { GLUCOSE, LACTATE, ACETATE, FORMATE,
               SUBSTANCE_COUNT } SubstanceId;

typedef struct _Substance
{
const char *name;
real32_t initValue,diffusivity,minDisplay,maxDisplay;
CudaBuffer gpu_c[3],gpu_dc[3];
CudaLockedMem mem_c[3],mem_dc[3];
real32_t *top_c[3],*top_dc[3];
} Substance;

typedef struct _Substrate
{
real32_t *planeDepths;
real32_t *planeSteps;
real32_t *masses[2];
CudaStream compStream,sendStream,recvStream;
CudaBuffer gpu_ref,gpu_tmp,gpu_diff;
CudaMarker compSync[3],sendSync[3],recvSync[3];
Substance substances[SUBSTANCE_COUNT];
int groupSize,groupCount;
CudaProgram rstPrg,refPrg,iterPrg;
#if ENABLE_RENDERING
  CudaBuffer gpu_tex;
  CudaProgram texPrg;
  int sides[6]; // W,E,S,N,D,U
#endif
} Substrate;

//----------------------------------------------------------------------------

#define MAX_REPLICATION_COUNT 2

struct _Bacterium
{
int8_t type; // minimize space
int8_t qGdev; // minimize space
int8_t replicationCount; // minimize space
uint8_t radius; // minimize space
real32_t x,y;
real32_t mass;
real32_t replicationProgress[MAX_REPLICATION_COUNT];
};

//----------------------------------------------------------------------------

typedef struct _DetectionCell
{
CACHE_ALIGNED // useless since explicitly aligned at allocation
real32_t roomLeft;
int count;
Bacterium content[1];
} DetectionCell;

//----------------------------------------------------------------------------

struct _DetectionCellStack
{
CACHE_ALIGNED
SpinSynchro sync;
int levelCount;
int64_t step;
intptr_t pageSize;
uint8_t *page;
};

//----------------------------------------------------------------------------

typedef struct _ActivationTile
{
CACHE_ALIGNED
Activity act;
SpinSynchro sync;
volatile bool empty;
} ActivationTile;

//----------------------------------------------------------------------------

struct _BacteriumStat
{
intptr_t count;
real64_t cG;
real64_t cL;
real64_t cA;
real64_t cF;
real64_t starving;
real64_t detection;
};

typedef struct _Stat
{
intptr_t nonEmptyTileCount;
intptr_t nonEmptyCellCount;
BacteriumStat bacteria[ASSUMED_MAX_BACTERIUM_TYPE_COUNT];
} Stat;

typedef struct _ThreadSimulationData
{
CACHE_ALIGNED
Stat stat;
} ThreadSimulationData;

#if ENABLE_RENDERING
  typedef enum { WEST_SIDE,  EAST_SIDE,
                 SOUTH_SIDE, NORTH_SIDE,
                 DOWN_SIDE,  UP_SIDE } SubstrateSide;
#endif

typedef struct _SimulationData
{
Scheduler sched;
bool mustQuit;
bool pause;
bool needsSubstrateReset;
#if ENABLE_RENDERING
  bool needsRedraw[2];
#endif
DetectionCellStack *detectionStacks[2];
ActivationTile *activationTiles;
Substrate substrate;
ThreadSimulationData tsd[ASSUMED_MAX_CPU_COUNT];
} SimulationData;

extern SimulationData sd;

//----------------------------------------------------------------------------

void
ConstantRecord_init(const char *fileName);

INLINE
const ConstantRecord *
ConstantRecord_get(void)
{
extern ConstantRecord _ConstantRecord_instance;
return &_ConstantRecord_instance;
}

#define IMPORT_STR_CONSTANT(cst,name) \
        const char * const cst##_##name=cst->name

#define IMPORT_INT_CONSTANT(cst,name) \
        const int cst##_##name=cst->name

#define IMPORT_R32_CONSTANT(cst,name) \
        const real32_t cst##_##name=cst->name

//----------------------------------------------------------------------------

void
Substrate_init(void);

void
Substrate_reset(void);

#if SYNCHRONISE_SUBSTRATE

  void
  Substrate_diffuse(void);

#else

  void
  Substrate_beginDiffusion(void);

  void
  Substrate_endDiffusion(void);

#endif

#if ENABLE_RENDERING
  void
  Substrate_updateTextureBuffer(int side,
                                int sideWidth,
                                int sideHeight,
                                int rendered);
#endif

//----------------------------------------------------------------------------

INLINE
real32_t
Bacterium_radiusFromMass(int type,
                         real32_t mass)
{
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
const BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+type;
IMPORT_R32_CONSTANT(bcst,mass_to_radius);
return cbrtf(mass*bcst_mass_to_radius);
}

INLINE
real32_t
Bacterium_cellOccupancyFromRadius(real32_t radius)
{
// return 3.4641f*radius*radius; // 2*sqrt(3) for hexagonal tiling
return M_PIf*radius*radius;
}

//----------------------------------------------------------------------------

void
ActivationTile_simulate(int cpuCtx,
                        void *data);

//----------------------------------------------------------------------------

#if ENABLE_RENDERING

void
Rendering_init(void);

int
Rendering_sideTextureBuffer(void);

void
Rendering_activate(int cpuCtx,
                   void *data);

void
Rendering_showBacterium(int cpuCtx,
                        int type,
                        real32_t x,
                        real32_t y,
                        real32_t radius,
                        int level,
                        real32_t starvation);

#endif

#endif // MICRODIMS_H

//----------------------------------------------------------------------------
